from __future__ import annotations

import dataclasses
import errno
import pathlib
import typing

import pysen
from pysen.component import LintComponentBase
from pysen.lint_command import LintCommandBase
from pysen.path import change_dir, resolve_path
from pysen.reporter import Reporter
from pysen.runner_options import PathContext
from pysen.setting import SettingBase, SettingFile, to_dash_case
from pysen.source import PythonFileFilter

from .preset import UNFIXABLE, preset_dict

CHECK_COMMAND = ["ruff", "check", "--show-source"]
FORMAT_COMMAND = ["ruff", "check", "--fix-only", "--show-fixes", "--exit-zero"]


class RuffCommand(LintCommandBase):
    def __init__(
        self: RuffCommand,
        paths: PathContext,
        source: pysen.Source,
        *,
        inplace_edit: bool,
    ) -> None:
        super().__init__(paths.base_dir, source)
        self._inplace_edit = inplace_edit
        self._setting_path = resolve_path(paths.settings_dir, "pyproject.toml")

    @property
    def name(self: RuffCommand) -> str:
        return "ruff"

    @property
    def has_side_effects(self: RuffCommand) -> bool:
        return self._inplace_edit

    def __call__(self: RuffCommand, reporter: Reporter) -> int:
        sources = self._get_sources(reporter, PythonFileFilter)
        msg = f"Checking {len(sources)} files"
        reporter.logger.info(msg)
        files = [str(s) for s in sources]
        command = FORMAT_COMMAND if self._inplace_edit else CHECK_COMMAND
        command = [*command, "--config", str(self._setting_path)]
        code = 0
        chunks = [files]
        with change_dir(self.base_dir):
            while len(chunks) > 0:
                chunk = chunks.pop()
                if len(chunk) == 0:
                    continue
                try:
                    returncode, _, _ = pysen.process_utils.run(
                        pysen.process_utils.add_python_executable(*command, *chunk),
                        reporter,
                    )
                    if returncode != 0:
                        code = returncode
                except OSError as e:
                    if len(chunk) >= 2 and e.errno == errno.E2BIG:
                        c = len(chunk) // 2
                        chunks.append(chunk[:c])
                        chunks.append(chunk[c:])
                    else:
                        raise

            return code


class RuffComponent(LintComponentBase):
    def __init__(
        self: RuffComponent,
        setting: RuffSetting,
        source: pysen.Source,
    ) -> None:
        super().__init__("ruff", source)
        self._setting = setting

    def export_settings(
        self: RuffComponent,
        paths: PathContext,
        files: typing.DefaultDict[str, SettingFile],
    ) -> None:
        files["pyproject.toml"].set_section(*self._setting.export())

    @property
    def targets(self: RuffComponent) -> typing.List[str]:
        return ["lint", "format"]

    def create_command(
        self: RuffComponent,
        target: str,
        paths: PathContext,
        options: pysen.RunOptions,
    ) -> pysen.CommandBase:
        return RuffCommand(paths, self.source, inplace_edit=(target == "format"))


class RuffPlugin(pysen.PluginBase):
    def load(
        self: RuffPlugin,
        file_path: pathlib.Path,
        config_data: pysen.PluginConfig,
        root: pysen.Config,
    ) -> typing.List[RuffComponent]:
        source = pysen.Source(includes=["."])
        setting = RuffSetting(**(config_data.config or {}))
        if root.lint:
            if root.lint.py_version:
                setting.target_version = root.lint.py_version.short_representation
            if root.lint.line_length:
                setting.line_length = root.lint.line_length
            if root.lint.isort_known_first_party:
                setting.isort = {"known-first-party": root.lint.isort_known_first_party}
            if root.lint.source:
                source = root.lint.source
        return [RuffComponent(setting, source=source)]


@dataclasses.dataclass(init=False)
class RuffSetting(SettingBase):
    select: typing.List[str]
    ignore: typing.List[str]
    unfixable: typing.List[str]
    target_version: typing.Optional[str] = None
    line_length: typing.Optional[int] = 88
    isort: typing.Optional[typing.Dict[str, typing.List[str]]] = None

    def __init__(self: RuffSetting, preset: str = "basic") -> None:
        self.select, self.ignore = preset_dict[preset]
        self.unfixable = UNFIXABLE

    def export(
        self: RuffSetting,
    ) -> typing.Tuple[typing.List[str], typing.Dict[str, typing.Any]]:
        return ["tool", "ruff"], self.asdict(omit_none=True, naming_rule=to_dash_case)



from .clang_format import clang_format  # NOQA
from .cmake_format import cmake_format  # NOQA
from .customizable_builtins import customizable_builtins  # NOQA
from .golint import golint  # NOQA
from .goreturns import goreturns  # NOQA
from .jq import jq  # NOQA
from .mypy_init_check import mypy_init_check  # NOQA
from .mypy_simple import mypy_simple  # NOQA
from .prettier import prettier  # NOQA
from .pylint import pylint  # NOQA
from .ruamel_yaml import ruamel_yaml  # NOQA
from .ruff import ruff  # NOQA
from .shellcheck import shellcheck  # NOQA
from .shfmt import shfmt  # NOQA
from .tidy import tidy  # NOQA



import dataclasses
import io
import pathlib
from typing import Optional, Sequence

import dacite

from pysen import (
    CommandBase,
    ComponentBase,
    Config,
    PluginBase,
    PluginConfig,
    RunOptions,
    SingleFileFormatCommandBase,
    Source,
)
from pysen.reporter import Reporter
from pysen.runner_options import PathContext
from pysen.types import TargetName

try:
    import ruamel.yaml

    _ruamel_yaml_import_error = None
except ImportError as e:
    _ruamel_yaml_import_error = e


def ruamel_yaml() -> PluginBase:
    if _ruamel_yaml_import_error:
        raise _ruamel_yaml_import_error
    return RuamelYamlPlugin()


class RuamelYamlPlugin(PluginBase):
    def load(
        self, file_path: pathlib.Path, config: PluginConfig, root: Config
    ) -> Sequence[ComponentBase]:
        if root.lint and root.lint.source:
            source = root.lint.source
        else:
            source = Source(includes=["."])

        ruamel_yaml_config = dacite.from_dict(
            RuamelYamlConfig,
            config.config or {},
            config=dacite.Config(strict=True),
        )
        assert isinstance(ruamel_yaml_config, RuamelYamlConfig)

        return [RuamelYamlComponent(source, ruamel_yaml_config)]


@dataclasses.dataclass
class RuamelYamlConfig:
    explicit_start: Optional[bool] = None
    preserve_quotes: Optional[bool] = None


class RuamelYamlComponent(ComponentBase):
    def __init__(self, source: Source, config: RuamelYamlConfig) -> None:
        self._source = source
        self._config = config

    @property
    def targets(self) -> Sequence[TargetName]:
        return ["lint", "format"]

    def create_command(
        self, target: str, paths: PathContext, options: RunOptions
    ) -> CommandBase:
        return RuamelYamlCommand(
            paths.base_dir, self._source, target == "format", self._config
        )


class RuamelYamlCommand(SingleFileFormatCommandBase):
    def __init__(
        self,
        base_dir: pathlib.Path,
        source: Source,
        inplace_edit: bool,
        config: RuamelYamlConfig,
    ) -> None:
        super().__init__(base_dir, source, inplace_edit)
        self._yaml = ruamel.yaml.YAML()
        self._yaml.explicit_start = config.explicit_start  # type: ignore
        self._yaml.preserve_quotes = config.preserve_quotes  # type: ignore

    @property
    def name(self) -> str:
        return "ruamel_yaml"

    @property
    def has_side_effects(self) -> bool:
        return self.inplace_edit

    def filter(self, file_path: pathlib.Path) -> bool:
        return file_path.suffix in {".yaml", ".yml"}

    def format(self, file_path: pathlib.Path, reporter: Reporter) -> Optional[str]:
        f = io.StringIO()
        try:
            self._yaml.dump(self._yaml.load(file_path), f)
        except ruamel.yaml.parser.ParserError as e:
            reporter.logger.exception(e)
            return None
        return f.getvalue()



import dataclasses
import logging
import pathlib
from typing import List, Optional, Sequence

import dacite

from pysen import (
    CommandBase,
    ComponentBase,
    Config,
    PluginBase,
    PluginConfig,
    RunOptions,
    SingleFileFormatCommandBase,
    Source,
    process_utils,
)
from pysen.reporter import Reporter
from pysen.runner_options import PathContext
from pysen.types import TargetName


def clang_format() -> PluginBase:
    return ClangFormatPlugin()


@dataclasses.dataclass
class ClangFormatConfig:
    extensions: Optional[List[str]] = None
    command: Optional[List[str]] = None


class ClangFormatPlugin(PluginBase):
    def load(
        self, file_path: pathlib.Path, config: PluginConfig, root: Config
    ) -> Sequence[ComponentBase]:
        if root.lint and root.lint.source:
            source = root.lint.source
        else:
            source = Source(includes=["."])

        plugin_config = dacite.from_dict(
            ClangFormatConfig,
            config.config or {},
            config=dacite.Config(strict=True),
        )
        assert isinstance(plugin_config, ClangFormatConfig)
        return [ClangFormatComponent(plugin_config, source)]


class ClangFormatComponent(ComponentBase):
    def __init__(self, config: ClangFormatConfig, source: Source) -> None:
        self._config = config
        self._source = source

    @property
    def targets(self) -> Sequence[TargetName]:
        return ["lint", "format"]

    def create_command(
        self, target: str, paths: PathContext, options: RunOptions
    ) -> CommandBase:
        return ClangFormatCommand(
            self._config, paths.base_dir, self._source, target == "format"
        )


class ClangFormatCommand(SingleFileFormatCommandBase):
    def __init__(
        self,
        config: ClangFormatConfig,
        base_dir: pathlib.Path,
        source: Source,
        inplace_edit: bool,
    ) -> None:
        super().__init__(base_dir, source, inplace_edit)

        self._config = config
        self._extensions = set(self._config.extensions or {".c", ".cpp", ".h", "hpp"})

    @property
    def name(self) -> str:
        return "clang_format"

    @property
    def has_side_effects(self) -> bool:
        return self.inplace_edit

    def filter(self, file_path: pathlib.Path) -> bool:
        return file_path.suffix in self._extensions

    def format(self, file_path: pathlib.Path, reporter: Reporter) -> Optional[str]:
        returncode, stdout, _ = process_utils.run(
            (self._config.command or ["clang-format"]) + [str(file_path)],
            reporter,
            stdout_loglevel=logging.NOTSET,
        )
        if returncode == 0:
            return stdout
        else:
            return None



import pathlib
from typing import Optional, Sequence

from pysen.command import CommandBase
from pysen.component import LintComponentBase
from pysen.flake8 import Flake8Command
from pysen.runner_options import PathContext, RunOptions
from pysen.source import Source


class Flake8(LintComponentBase):
    def __init__(
        self,
        name: str = "flake8",
        source: Optional[Source] = None,
        settings_dir: Optional[pathlib.Path] = None,
    ) -> None:
        super().__init__(name, source)
        self._settings_dir = settings_dir or pathlib.Path(".")

    @property
    def targets(self) -> Sequence[str]:
        return ["lint"]

    def create_command(
        self, target: str, paths: PathContext, options: RunOptions
    ) -> CommandBase:
        paths = PathContext(base_dir=paths.base_dir, settings_dir=self._settings_dir)
        if target == "lint":
            return Flake8Command(self.name, paths, self.source)

        raise AssertionError(f"unknown {target}")



import dataclasses
import pathlib
from typing import Any, Dict, List, Optional, Sequence

import dacite

from pysen import PluginBase
from pysen.component import ComponentBase
from pysen.exceptions import InvalidConfigurationError
from pysen.mypy import MypyTarget
from pysen.pyproject_model import (
    Config,
    PluginConfig,
    _expand_path,
    _parse_mypy_targets,
    _parse_source,
)
from pysen.source import Source
from pysen_plugins.customizable_builtins.black import Black
from pysen_plugins.customizable_builtins.flake8 import Flake8
from pysen_plugins.customizable_builtins.isort import Isort
from pysen_plugins.customizable_builtins.mypy import Mypy


@dataclasses.dataclass
class CustomizablePysenConfigureLintOptions:
    enable_black: Optional[bool] = None
    enable_flake8: Optional[bool] = None
    enable_isort: Optional[bool] = None
    enable_mypy: Optional[bool] = None
    source: Optional[Source] = None
    mypy_targets: Optional[List[MypyTarget]] = None


@dataclasses.dataclass
class CustomizablePysenConfig:
    lint: Optional[CustomizablePysenConfigureLintOptions]
    settings_dir: Optional[pathlib.Path]


def _parse_dict(
    data: Dict[str, Any], base_dir: pathlib.Path
) -> CustomizablePysenConfig:
    dacite_config = dacite.Config(
        type_hooks={
            Source: lambda x: _parse_source(base_dir, x),
            pathlib.Path: lambda x: _expand_path(base_dir, x),
            List[MypyTarget]: lambda x: _parse_mypy_targets(base_dir, x),
        },
        strict=True,
    )

    try:
        config = dacite.from_dict(CustomizablePysenConfig, data, dacite_config)
        assert isinstance(config, CustomizablePysenConfig)
        return config
    except dacite.DaciteError as e:
        raise InvalidConfigurationError(f"invalid configuration: {e}") from None


class CustomizablePysenPlugin(PluginBase):
    def load(
        self, file_path: pathlib.Path, config: PluginConfig, root: Config
    ) -> Sequence[ComponentBase]:
        base_dir = file_path.parent
        plugin_config = _parse_dict(data=config.config or {}, base_dir=base_dir)
        lint_config = plugin_config.lint or CustomizablePysenConfigureLintOptions()
        components: List[ComponentBase] = []
        if lint_config.enable_black:
            components.append(
                Black(
                    source=lint_config.source,
                    settings_dir=plugin_config.settings_dir or base_dir,
                )
            )
        if lint_config.enable_flake8:
            components.append(
                Flake8(
                    source=lint_config.source,
                    settings_dir=plugin_config.settings_dir or base_dir,
                )
            )
        if lint_config.enable_isort:
            components.append(
                Isort(
                    source=lint_config.source,
                    settings_dir=plugin_config.settings_dir or base_dir,
                )
            )
        if lint_config.enable_mypy:
            components.append(
                Mypy(
                    mypy_targets=lint_config.mypy_targets,
                    settings_dir=plugin_config.settings_dir or base_dir,
                )
            )

        return components


def customizable_builtins() -> CustomizablePysenPlugin:
    return CustomizablePysenPlugin()



import pathlib

from pysen import Source, process_utils
from pysen.error_lines import parse_error_lines
from pysen.lint_command import LintCommandBase
from pysen.path import change_dir
from pysen.reporter import Reporter
from pysen.source import PythonFileFilter


class PylintCommand(LintCommandBase):
    def __init__(
        self, base_dir: pathlib.Path, source: Source, settings_path: pathlib.Path
    ) -> None:
        super().__init__(base_dir, source)
        self._source = source
        self._base_dir = base_dir

    @property
    def name(self) -> str:
        return "pylint"

    @property
    def has_side_effects(self) -> bool:
        return False

    def __call__(self, reporter: Reporter) -> int:
        sources = self._source.resolve_files(
            self._base_dir, PythonFileFilter, reporter=reporter
        )
        cmd = [
            "pylint",
            "--msg-template='{abspath}:{line}:{column}: {msg}'",
            "--score=n",
        ]
        cmd.extend(map(str, sources))
        with change_dir(self.base_dir):
            ret, stdout, _ = process_utils.run(cmd, reporter)

        # NOTE(ryo): We cannot control the output of pylint to show only
        # parseable lines, so we optimistically use "/" to determine whether
        # the line contains a parseable error.
        #
        # Example output:
        # ************* Module source
        # /examples/pylint/source.py:1:0: Missing module docstring
        # ************* Module hoge
        # /examples/pylint/hoge.py:1:0: Similar lines in 2 files
        # ==foo:0
        # ==bar:0
        # Class MyClass:
        #     def __init__(self):
        #         self._a = 1
        stdout = "\n".join(line for line in stdout.splitlines() if line.startswith("/"))
        diagnostics = parse_error_lines(stdout, logger=reporter.logger)
        reporter.report_diagnostics(list(diagnostics))

        return ret
